import rich_click as click

from novara.utils import logger

@click.command()
def version():
    """Display the current version of the CLI."""
    from novara.constants import __version__

    logger.info(f"Current CLI version: {__version__}")
    logger.info("For more information, visit https://github.com/TeamM0unt41n/novara")
    logger.info("For help, use the command: novara --help")


import os
import toml
import json

import rich_click as click

from questionary import text, confirm

from novara.utils import logger
from novara.request import request, JSONDecodeError
from novara.guess import guess_pip_packages


def generate_dockerfile():
    try:
        with open("novara.toml", "r") as f:
            toml_raw = f.read()
    except (OSError, FileNotFoundError):
        raise click.ClickException("novara.toml either not there or unaccessable")

    logger.info("read toml file")

    logger.info("generating Dockerfile")
    r = request.post(
        f"api/build/dockerfile/toml/", data=json.dumps(toml_raw)
    )
    if not r.ok:
        raise click.ClickException(
            f"failed generating dockerfile with message: {r.text}"
        )
        exit()

    try:
        dockerfile_str = r.json()
    except JSONDecodeError:
        raise click.ClickException(
            f"failed to decode response as json: {r.text[:20] if len(r.text) > 20 else r.text}"
        )

    try:
        with open("Dockerfile", "w") as docker_file:
            docker_file.write(dockerfile_str)
    except OSError:
        raise click.ClickException("Dockerfile is unaccessable")
        exit()

    logger.info("generated Dockerfile")
    logger.info("run 'novara run local|remote' to start a container or remote container")


@click.command()
@click.option("--pip", default=None, help="add a new pip packages")
@click.option("--apt", default=None, help="add a new apt packages")
@click.option("--apk", default=None, help="add a new apk packages")
@click.option("--snap", default=None, help="add a new snap packages")
@click.option("--dnf", default=None, help="add a new dnf packages")
@click.option("--no-guess", default=False, help='disable guessing the necessary python packages', is_flag=True)
@click.option("--directory", "-d", default="./", help="path to exploit")
def generate(pip, apt, apk, snap, dnf, no_guess, directory):
    """(re)generate the Dockerfile and add dependencies if necessary"""

    directory = os.path.join(directory, "novara.toml")

    if any([pip, apt, apk, snap, dnf]):
        logger.info("adding packages to novara.toml file")

    try:
        with open(directory, "r") as f:
            exploit_config:dict[str, dict[str, dict[str, list[str]]]] = toml.load(f)
    except (OSError, FileNotFoundError):
        raise click.ClickException(f"{directory} either not there or unaccessable")

    docker_config = exploit_config.get("docker", None)
    if docker_config is None:
        raise click.ClickException("docker key not found in toml")
        
    dependencies = docker_config.get("dependencies", None)
    if dependencies is None:
        raise click.ClickException("docker.dependencies key not found in toml")

    pip_packages = dependencies.get("pip", None)
    apt_packages = dependencies.get("apt", None)
    apk_packages = dependencies.get("apk", None)
    dnf_packages = dependencies.get("dnf", None)
    snap_packages = dependencies.get("snap", None)

    guessed_packages = []
    if not no_guess:
        guessed_packages = guess_pip_packages(exploit_config.get('script_file_name', 'exploit.py'), exclude=pip_packages or [])

        if guessed_packages:
            logger.info(f'Found packages: {", ".join(guessed_packages)}')

            if confirm(f'Do you want to add these packages ({", ".join(guessed_packages)})?', default=True).ask():
                for pkg in guessed_packages:
                    package = text('Is the name/version of this package correct (to omit just delete name)?', default=pkg).ask()
                    if not package:
                        continue
                    if not pip_packages:
                        pip_packages = []
                    pip_packages.append(package)

    if pip is not None:
        dependencies.update(
            {"pip": [pip] if pip_packages is None else pip_packages + [pip] }
        )
    
    if apt is not None:
        dependencies.update(
            {"apt": [apt] if apt_packages is None else apt_packages + [apt]}
        )
    
    if apk is not None:
        dependencies.update(
            {"apk": [apk] if apk_packages is None else apk_packages + [apk]}
        )
    
    if dnf is not None:
        dependencies.update(
            {"dnf": [dnf] if dnf_packages is None else dnf_packages + [dnf]}
        )
    
    if snap is not None:
        dependencies.update(
            {"snap": [snap] if snap_packages is None else snap_packages + [snap]}
        )
    
    try:
        with open(directory, "w") as f:
            f.write(
                "# novara configuration file autogenerated by the novara cli, only edit if you know what you are doing\n\n"
            )
            toml.dump(exploit_config, f)
    
    except (OSError, FileNotFoundError):
        raise click.ClickException(f"{directory} either not there or unaccessable")
    
    logger.info("updated novara.toml file")

    generate_dockerfile()



import ast
import sys
import importlib.util
import importlib.metadata

def resolve_import(import_name) -> str:
    try:
        packages = importlib.metadata.packages_distributions()
        distributions = packages.get(import_name, [])
        return distributions[0] if distributions else import_name
    except Exception:
        return import_name

def resolve_package_version(package_name:str) -> None | str:
    try:
        version = importlib.metadata.version(package_name)
        return version
    except importlib.metadata.PackageNotFoundError:
        return None
    
def is_builtin_module(module_name:str) -> bool:
    if module_name in sys.builtin_module_names:
        return True
    try:
        spec = importlib.util.find_spec(module_name)
        if spec is None:
            return False
        if spec.origin == 'built-in':
            return True
        if 'site-packages' in spec.origin or 'dist-packages' in spec.origin:
            return False
        return True
    except ModuleNotFoundError:
        return False


def guess_pip_packages(file:str, exclude:list[str] = []) -> list[str]:
    with open(file) as f:
        python_code = f.read()

    tree = ast.parse(python_code)

    imports:list[str] = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for direct_import in node.names:
                imports.append(direct_import.name)
        elif isinstance(node, ast.ImportFrom):
            imports.append(node.module)

    packages = []

    imports = set( [imprt.split('.', maxsplit=1)[0] for imprt in imports])

    for imprt in imports:
        package = resolve_import(imprt)
        if is_builtin_module(package) or any(package == exclusion.split('==', maxsplit=1)[0] for exclusion in exclude):
            continue

        version = resolve_package_version(package)
        if version is not None:
            package += f'=={version}'

        packages.append(package)
 
    return packages


import rich_click as click
from novara.utils import logger
import os
import toml
from novara.request import request
from questionary import confirm


@click.command()
@click.option("-f","--force", default=False, is_flag=True, help="skip prompt")
@click.option("--directory", "-d", default="./", help="path to exploit")
def stop(directory, force):
    """stop the current running exploit"""
    
    directory = os.path.join(directory, "novara.toml")

    try:
        with open(directory, "r") as f:
            exploit_config = toml.load(f)
    except (OSError, FileNotFoundError):
        raise click.ClickException(f"{directory} either not there or unaccessable")

    if not force:
        confimation = confirm(f'Do you really want to stop exploit {exploit_config["service"]["name"]}-{exploit_config["exploit_name"]}').ask()
        if not confimation:
            logger.info('canceled stopping of exploit')
            exit()
    
    exploit_id = exploit_config["exploit_id"]

    logger.info("stopping container...")
    r = request.delete(f"api/exploits/container/{exploit_id}/")
    if not r.ok:
        raise click.ClickException(
            f"failed stopping container with message: {r.text}"
        )
    
    logger.info(f'Stopped exploit {exploit_config["service"]["name"]}-{exploit_config["exploit_name"]}')


from rich.console import Console
from rich.logging import RichHandler
from random import randrange, seed
import logging
import os
from novara.config import config
from novara.constants import SSHKEY_FILE
from novara.request import AuthSession
import paramiko

# -----------------------------------------------------------------

logging.basicConfig(
    level=config.logging_level if config.is_initialized else 'NOTSET',
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)

logger = logging.getLogger("rich")
console = Console()

# -----------------------------------------------------------------


def print(*args, **kwargs):
    console.print(*args, **kwargs)


# -----------------------------------------------------------------


def color_value(value: str):
    seed(value.lower())
    r, g, b = [str(hex(randrange(25, 255))[2:]) for _ in range(3)]
    value_colored = f"[bold #{r}{g}{b}]{value}[/]"

    return value_colored


# -----------------------------------------------------------------

def test_ssh_connection():
    # --------------
    try:
        with open(SSHKEY_FILE, "w") as f:
            f.write(config.ssh_privatekey)
    except OSError:
        logger.error("Couldn't create the SSH-key it's not writable")
        exit()
    # --------------
    try:
        os.chmod(SSHKEY_FILE, 0o600)
    except OSError:
        logger.error("Couldn't change the SSH-key's permissions")
        exit()
    # --------------
    logger.info("Testing SSH connection...")

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        private_key = paramiko.RSAKey.from_private_key_file(SSHKEY_FILE)
        ssh.connect(
            hostname=config.ssh_url,
            port=config.ssh_port,
            username=config.ssh_user,
            pkey=private_key,
            timeout=10
        )
        stdin, stdout, stderr = ssh.exec_command("echo :3")
        error_output = stderr.read().decode().strip()

        # Check if there was any error output
        if error_output:
            logger.error("Failed to establish SSH connection!!!")
            logger.debug("Is the public key on the target's authorized_keys file?")
            logger.error(error_output)
            return False

        logger.info("SSH connection successful")
        return True

    except paramiko.SSHException as e:
        logger.error("Failed to establish SSH connection!!!")
        logger.error(f"SSHException: {e}")
        return False

    except Exception as e:
        logger.error("An unexpected error occurred while trying to connect via SSH.")
        logger.error(f"Exception: {e}")
        return False

    finally:
        ssh.close()


import rich_click as click
from urllib.parse import urljoin, urlparse
import os
import requests
import logging
from novara.request import AuthSession
from novara.config import config, ConfigManager
from novara.utils import test_ssh_connection

from auto_click_auto import enable_click_shell_completion  

logger = logging.getLogger('rich')


@click.command()
@click.option(
    "--server_url",
    "-l",
    default=None,
    help="url to api-endpoint of your novara instance",
)
@click.option(
    "--author",
    "-a",
    default=None,
    help="to specify what author to use for the exploits",
)
@click.option(
    "--log-level",
    default="NOTSET",
    help="Change the log level of the logger (CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET)"
)
@click.option(
    "--auto-complete",
    is_flag=True,
    help="Enable auto-completion for the CLI commands",
)
def configure(server_url:str, author:str, log_level:str, auto_complete:bool):
    """conect to novara backend & configure the cli"""

    log_level = log_level.upper()

    if log_level not in ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET']:
        raise click.ClickException(f"invalid log level: {log_level}, must be one of: CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET")

    # Priority: CLI argument > Environment variable > Prompt

    logging.basicConfig(level=log_level)

    server_url = (
        server_url
        or os.environ.get("SERVER_URL")
        or click.prompt("Please enter the Novara server URL")
    )
    try:
        r = requests.get(urljoin(server_url, '/api/config/auth_config/'))
        if not r.ok:
            raise click.ClickException(f"the remote responded with error:\n{r.text}")
        
        local_config = ConfigManager.model_validate(r.json() | {'server_url': server_url, 'logging_level':log_level})

    except requests.JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")

    except requests.exceptions.ConnectionError:
        logger.error(f'failed to connect to backend. Is the url correct?')
        exit()

    
    session = AuthSession(local_config)

    userinfo = session.get_userinfo()

    if userinfo is None:
        logger.warning("failed to retrieve userinfo, please try again")

    local_config.author = (
        author
        or os.environ.get("AUTHOR_NAME")
        or (userinfo.preferred_username if userinfo else None)
        or (userinfo.given_name if userinfo else None)
        or click.prompt("Please enter your author username")
    )

    print('session configured')

    # -----------------------------------------------------------------

    r = session.get(urljoin(local_config.server_url, "/api/config/cli/"),
        params={'username':author}
    )
    if not r.ok:
        raise click.ClickException(f"the remote responded with error:\n{r.text}")

    # -----------------------------------------------------------------

    try:
        local_config.update(**r.json())
    except requests.JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")
    
    if auto_complete:
        enable_click_shell_completion('novara', verbose=log_level.lower() != 'debug')


from rich import traceback

# traceback.install(show_locals=True)

import rich_click as click
from novara.commands.init import init
from novara.commands.status import status
from novara.commands.configure import configure
from novara.commands.pull import pull
from novara.commands.docker import docker
from novara.commands.generate import generate
from novara.commands.run import run
from novara.commands.logs import logs
from novara.commands.stop import stop
from novara.commands.remove import remove
from novara.commands.info import info 
from novara.commands.version import version


@click.group()
def main():
    """novara is a cli tool to help in A/D CTFs"""


main.add_command(init)
main.add_command(configure)
main.add_command(status)
main.add_command(pull)
main.add_command(generate)
main.add_command(run)
main.add_command(logs)
main.add_command(stop)
main.add_command(remove)
main.add_command(info)
main.add_command(version)


import rich_click as click
import subprocess
import os
import time
from novara.utils import logger
from novara.constants import SSHKEY_FILE, SOCKET_FILE
from novara.config import config


def forward_docker_socket():
    if os.path.exists(SOCKET_FILE):
        os.remove(SOCKET_FILE)

    return subprocess.Popen(["ssh", "-i", SSHKEY_FILE, f"{config.ssh_user}@{config.ssh_url}", "-p", f"{config.ssh_port}", "-L", f"{SOCKET_FILE}:/var/run/docker.sock", "--", "sleep", "infinity"], stdout=subprocess.DEVNULL, stdin=subprocess.DEVNULL)

def cleanup_docker_socket(ssh:subprocess.Popen):
    ssh.kill()

    if os.path.exists(SOCKET_FILE):
        os.remove(SOCKET_FILE)


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.pass_context
def docker(ctx):
    """this command allows you to control the docker on the remote with the docker cli"""

    ssh = forward_docker_socket()
    
    env = {
        **os.environ.copy(),
        "DOCKER_HOST": f"unix://{SOCKET_FILE}",
    }

    try:
        while not os.path.exists(SOCKET_FILE):
            time.sleep(0.1)

        docker = subprocess.Popen(["docker", *ctx.args], env=env)
        
        docker.wait()
    except KeyboardInterrupt:
        logger.warning("terminating cli")
    docker.kill()
    cleanup_docker_socket(ssh)


import rich_click as click
from novara.utils import logger
from novara.commands.generate import generate_dockerfile
import docker
import toml
from io import BytesIO
from zipfile import ZipFile
import os
import sys
from novara.request import request, JSONDecodeError
import json
import stat
import docker.models.containers
import subprocess
from docker.errors import BuildError

os_windows = os.name == "nt"

def build_image(*args, **kwargs):
    client = docker.from_env()

    kwargs.update({'decode':True, 'pull':True})

    response_stream:list[dict[str, str]] = client.api.build(*args, **kwargs)

    status_dict = {}

    for chunk in response_stream:
        type, *other_info = chunk.keys()
        match type:
            case 'error':
                raise BuildError(chunk['error'], response_stream)
            case 'errorDetail':
                sys.stdout.write(chunk)
            case 'stream':
                sys.stdout.write(chunk['stream'])
            case 'status':
                if 'id' in other_info:
                    change_id = chunk['id']
                    if not change_id in status_dict:
                        sys.stdout.write('\n')
                    status_dict.update({change_id:{'id':change_id, 'status':chunk['status'], 'progress':chunk.get('progress', '')}})
                    for _ in status_dict.keys():
                        sys.stdout.write('\033[F\033[K')
                    for update in status_dict.values():
                        sys.stdout.write(f"{update['id']}: {update['status']} {update.get('progress', '')}\n")
                else:
                    sys.stdout.write(chunk['status'])
            case 'aux':
                value = chunk['aux']
                if 'ID' in value:
                    image_id = value['ID'].split(':')[-1]
                else:
                    sys.stdout.write(f'{chunk}')
            case _:
                sys.stdout.write(f'{chunk}')

    image = None

    if image_id:
        image = client.images.get(image_id)

    return image

@click.group()
def run():
    """run a exploit script either locally or on remote"""

@run.command()
def shell():
    """run the exploit in your current shell"""
    try:
        with open("novara.toml", "r") as f:
            config_toml = toml.load(f)
    except (OSError, FileNotFoundError):
        raise click.ClickException("novara.toml either not there or unaccessable")
    if not 'sploit' in config_toml:
        raise click.ClickException("sploit needs to be specified in toml")
    if not 'script_name' in config_toml['sploit']:
        raise click.ClickException("sploit.script_name needs to be specified in toml")
    if not 'args' in config_toml['sploit']:
        raise click.ClickException("sploit.args needs to be specified in toml") 
    if not 'script_file' in config_toml:
        raise click.ClickException("script_name needs to be specified in toml") 

    start_sploit_script_path = os.path.join(os.getcwd(), config_toml['sploit']['script_name'])

    if not os_windows:
        file_mode = os.stat(start_sploit_script_path).st_mode
        if not file_mode & stat.S_IXUSR:
            logger.info("Setting the executable bit on `{}`".format(start_sploit_script_path))
            os.chmod(start_sploit_script_path, file_mode | stat.S_IXUSR)

    exploit_process = subprocess.Popen([os.path.join(os.getcwd(), config_toml['sploit']['script_name']), *config_toml['sploit']['args'], os.path.join(os.getcwd(), config_toml['script_file'])])

    try:
        exploit_process.wait()
    except KeyboardInterrupt:
        logger.warning("terminating exploit")
    exploit_process.kill()

@run.command()
def local():
    """run the exploit script locally"""
    client = docker.from_env()

    generate_dockerfile()

    logger.info("building image")
    image = build_image(path='.')

    logger.info("starting container")
    container: docker.models.containers.Container = client.containers.run(
        image, detach=True
    )
    try:
        for docker_log in container.logs(stream=True):
            click.echo(docker_log, nl=False)
    except KeyboardInterrupt:
        logger.warning("stopping container")
        container.kill()


@run.command()
@click.option(
    "-k",
    "--keep",
    default=False,
    is_flag=True,
    help="keep container, if it already exists",
)
@click.option(
    '-g',
    '--generate',
    default=False,
    is_flag=True,
    help="generate the dockerfile from the included toml file",
)
def remote(keep:bool, generate:bool):
    """upload and run the exploit on the remote server"""

    try:
        with open("novara.toml", "r") as f:
            toml_config = toml.load(f)
    except (OSError, FileNotFoundError):
        raise click.ClickException("novara.toml either not there or unaccessable")
    logger.info("read toml file")

    if "exploit_id" not in toml_config:
        raise click.ClickException(
            "no exploit_id in toml found, consider regenerating the toml with novara init"
        )

    exploit_id = toml_config.get("exploit_id")

    if not exploit_id:
        raise click.ClickException(
            "exploit_id is empty, consider regenerating the toml with novara init"
        )

    zip_archive = BytesIO()

    with ZipFile(zip_archive, "w") as zip:
        for root, _, filenames in os.walk(os.getcwd()):
            for name in filenames:
                zip.write(name, name)

    zip_archive.seek(0)

    logger.info("uploading zip...")

    r = request.post(f"api/exploits/{exploit_id}/", files={"file": zip_archive})
    if not r.ok:
        raise click.ClickException(f"Uploading zip failed with error: {r.text}")
        exit()

    if os.path.exists('Dockerfile') and not generate:
        try:
            with open("Dockerfile", "r") as f:
                docker_file = f.read()
        except (OSError, FileNotFoundError):
            raise click.ClickException("Dockerfile unaccessable")
        logger.info("read toml dockerfile")

        logger.warning('Using dockerfile to build container, use --generate to force rebuild from toml.')
        r = request.post(f"api/build/{exploit_id}/docker/", data=json.dumps(docker_file), stream=True)
        if not r.ok:
            raise click.ClickException(
                f"Failed building dockerfile file with error: {r.text}"
            )
        content_iterator = r.iter_lines()
        for line_str in content_iterator:
            line = json.loads(line_str)
            match line['type']:
                case 'logs':
                    sys.stdout.write(line['data'])
                case 'image':
                    image = line['data']
    else:
        try:
            with open("novara.toml", "r") as toml_file:
                toml_str = toml_file.read()
        except (FileNotFoundError, OSError):
            raise click.ClickException("Failed reading novara toml file")

        logger.info("building image...")

        r = request.post(f"api/build/{exploit_id}/toml/", data=json.dumps(toml_str), stream=True)
        if not r.ok:
            raise click.ClickException(
                f"Failed building novara toml file with error: {r.text}"
            )
            exit()

        logger.info(f'build logs:')

        content_iterator = r.iter_lines()
        for line_str in content_iterator:
            line = json.loads(line_str)
            match line['type']:
                case 'logs':
                    sys.stdout.write(line['data'])
                case 'image':
                    image = line['data']
    
    if image is None:
        raise click.ClickException("Something went wrong while building image")
        exit()

    logger.info(f"image: {image}")

    if not keep:
        logger.info(f"requesting removal of old containers of exploit: {exploit_id}")
        r = request.delete(f"api/exploits/container/{exploit_id}/")
        if r.status_code == 404:
            logger.info("No containers found for current exploit")
        if not r.ok:
            logger.warning(f"failed removing container with message: {r.text}")

    logger.info("starting new container...")

    r = request.post("api/containers/", params={"exploit_id": exploit_id})
    if not r.ok:
        raise click.ClickException(
            f"Failed starting new container with error: {r.text}"
        )
        exit()

    try:
        container_info = r.json()
    except JSONDecodeError:
        raise click.ClickException(
            f"failed to decode response as json: {r.text[:20] if len(r.text) > 20 else r.text}"
        )
    for info in container_info:
        logger.info(f"{info}: {container_info[info]}")

    logger.info(
        "Done deploying new container, to redeploy the container just run 'novara run remote' again"
    )



import rich_click as click
from novara.request import request, JSONDecodeError
from novara.utils import logger
from questionary import confirm, select
from io import BytesIO
from zipfile import ZipFile
import os
import shutil
import toml

def delete_all_in_directory(directory_path):
    for filename in os.listdir(directory_path):
        file_path = os.path.join(directory_path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.remove(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)

def download_empty_template(exploit_id:str):

    r = request.get('api/exploits/')
    if not r.ok:
        raise click.ClickException(
            f"Requesting exploits from remote failed with error: {r.text}. Did you run novara configure?"
        )
    
    try:
        exploits = r.json()
    except JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")
    
    service = None

    for exploit in exploits:
        if exploit['id'] == exploit_id:
            service = exploit['service']

    if service is None:
        raise click.ClickException(f'exploit with id {exploit_id} not found on remote')

    r = request.post(
        f"api/services/{service}/template/",
        params={"exploit_id": exploit_id, "exploit_name":'', "exploit_author": ""},
    )
    if not r.ok:
        raise click.ClickException(
            f"Requesting template from remote failed with error: {r.text}. Did you run novara configure?"
        )
    return r

def download_exploit(exploit_id: str, exploit_name: str, dir:str, files:list[str], is_current_dir:bool=False):
    version = -1

    if files is not None:
        if len(files):
            choices = [str(x+1) for x in range(len(files))]
            choices[0] = f'{choices[0]} (initial)'
            choices[-1] = f'{choices[-1]} (latest)'

            selected = select('Please select a version:', choices=choices[::-1]).ask()
            version = choices.index(selected)
        else:
            logger.warning('no version found')

    if dir is None:
        dir = os.path.normpath(os.path.join(os.getcwd(), exploit_name))
    logger.info(f"exploit id: {exploit_id}")
    logger.info("requesting exploit")

    r = request.get(f"api/exploits/{exploit_id}/", params={'version':version})
    if not r.ok:
        if 'no files' in r.text:
            logger.info("The requested exploit doesn't yet have any files yet.")
            result = confirm('Do you want to download a empty template?').ask()
            if result:
                r = download_empty_template(exploit_id)
        
        else:
            raise click.ClickException(
                f"Requesting exploit from remote failed with error: {r.text}. Did you run novara configure?"
            )
    logger.info("extracting exploit")
    zip_template = BytesIO(r.content)
    if os.path.exists(dir):
        overwrite = is_current_dir or confirm(
            message=f"Do you want to overwrite the directory '{dir}'? (Y/n)",
            default=True
        ).ask()
        if not overwrite:
            raise click.ClickException(
                "Directory already exists and can't be overwritten. Consider using the -d or --dir to specify a different directory."
            )
            exit()
        delete_all_in_directory(dir)
    else:
        os.mkdir(dir)
    with ZipFile(zip_template) as zip:
        zip.extractall(dir)

    logger.info("Exploit extracted sucessfully")


@click.command()
@click.option(
    "-s", "--service", default=None, help="name of the service the exploit is targeting"
)
@click.option("-n", "--name", default=None, help="full name of the exploit to pull")
@click.option(
    "-d",
    "--directory",
    default=None,
    help="specify a different directory to save the exploit",
)
@click.option("-l", "--listall", default=False, help="list all exploits", is_flag=True)
@click.option("-v", "--version", default=False, help="select a version (pulls latest version by default)", is_flag=True)
@click.argument("pull_name", required=False)
def pull(service, name, directory, listall, version, pull_name):
    """Pull existing exploits from novara, if no other arguments are provided and there is a novara.toml in the same directory the cli will try to pull the newest version of the exploit"""
    
    r = request.get("api/exploits/")
    if not r.ok:
        raise click.ClickException(
            f"failed requesting a list of exploits from remote with error: {r.text}"
        )
    try:
        exploits:list[dict] = r.json()
    except JSONDecodeError:
        raise click.ClickException(f"unable to decode response as json:\n{r.text}")
    
    name = name or pull_name

    exploit = None
    exploit_config = None
    if service is None and name is None and directory is None and not listall:
        try:
            with open("novara.toml", "r") as f:
                exploit_config = toml.load(f)
        except (OSError, FileNotFoundError):
            listall = True
    if not exploit_config is None:
        exploit_id = exploit_config.get('exploit_id')
        for e in exploits:
            if e['id'] == exploit_id:
                exploit = e
                break

    if exploit is not None:         
        download_exploit(exploit['id'], f"{exploit['service']}-{exploit['name']}", os.getcwd(), files=exploit.get('files', []) if version else None, is_current_dir=True)
        exit()

    # -----------------------------------------------------------------

    # check if any exploits are even available
    if not len(exploits):
        raise click.ClickException("No exploits to pull!")

    # pull by name ----------------------------------------------------
    if name and any(exploit["name"] == name or f'{exploit["service"]}-{exploit["name"]}' == name for exploit in exploits):
        exploit = None
        for exploit in exploits:
            if exploit["name"] == name or f'{exploit["service"]}-{exploit["name"]}' == name:
                exploit = exploit
                break
        download_exploit(exploit["id"], name, directory, files=exploit.get('files', []) if version else None)
        exit()
    elif name:
        raise click.ClickException(f"No exploits named '{name}' available!")

    # chose a service -------------------------------------------------

    if listall:
        # pull by selection ----------------------------------------
        identifier = select(
            "Chose a exploit:",
            choices=[f'{e["service"]}-{e["name"]}' for e in exploits],
        ).ask()

        if identifier is None:
            exit()

        for exploit in exploits:
            if f'{exploit["service"]}-{exploit["name"]}' == identifier:
                download_exploit(exploit['id'], identifier, directory, files=exploit.get('files', []) if version else None)
                exit()
    
    elif service is None:
        services = []
        for exploit in exploits:
            if not exploit['service'] in services:
                services.append(exploit['service'])
        service = select(
                "Please select a service:",
                choices=services,
            ).ask()
        
        if service is None:
            logger.warning('canceling pull')
            return

    # pull by service -------------------------------------------------
    if service and any(exploit["service"] == service for exploit in exploits):
        service_exploits:dict[dict] = {}
        for exploit in exploits:
            if exploit["service"] == service:
                service_exploits[exploit["name"]] = exploit

        if len(list(service_exploits.keys())) > 1:
            service = select(
                "there are multiple exploits available for this service, please select one:",
                choices=list(service_exploits.keys()),
            ).ask()
            if service is None:
                logger.warning('canceling pull')
                return
            exploit = service_exploits[service]
        else:
            exploit = list(service_exploits.values())[0] 
        
        download_exploit(exploit["id"], f"{exploit['service']}-{exploit['name']}", directory, files=exploit.get('files', []) if version else None)
        exit()

    elif name:
        raise click.ClickException(f"No exploits for '{service}' available")



import rich_click as click
import os
import shutil
import stat

from io import BytesIO
from questionary import confirm, select, path
from zipfile import ZipFile

from novara.config import config
from novara.utils import logger
from novara.request import request, JSONDecodeError

def get_service():
    r = request.get("api/services/")
    if not r.ok:
        raise click.ClickException(
            f"Failed requesting list of services from remote with error: {r.text}"
        )

    try:
        services = r.json()
    except JSONDecodeError:
        raise click.ClickException(
            f"failed to decode response as json: {r.text[:20] if len(r.text) > 20 else r.text}"
        )

    service = select("Please select a service", choices=services).ask()

    if service is None:
        logger.warning('Cancelled by user, exiting...')
        exit()

    return service


@click.command()
@click.option(
    "-s",
    "--service",
    default=None,
    help="the name of the service the exploit will be attacking",
)
@click.option(
    "-n", "--name", default=None, help="the internal name for the exploit identifing it"
)
@click.option('-t', '--template-type', default=None, help=f'the type of the template', )
@click.option("-a", "--author", default=None, help="name of the exploit's author")
@click.option(
    "-d",
    "--directory",
    default=None,
    help="specify a different directory to put the exploit",
)
def init(service, name, template_type, author, directory):
    """Initialize a new exploit from a template"""

    # Priority: CLI argument > Environment variable > Prompt

    r = request.get('api/exploits/generate_name')
    if not r.ok:
        raise click.ClickException(
            f"Requesting name from remote failed with error: {r.text}."
        )
    
    name = name or r.json()

    
    r = request.get('api/services/template/types')
    if not r.ok:
        raise click.ClickException(
            f"Requesting template types from remote failed with error: {r.text}."
        )
    
    valide_template_types = r.json()

    service = service or get_service()
    author = (
        author
        or config.author
        or os.environ.get("AUTHOR")
        or click.prompt("Please enter this exploit author's name")
    )

    if author is None:
        logger.warning('Cancelled by user, exiting...')
        exit()

    type = (
        template_type 
        or os.environ.get("TEMPLATE-TYPE")
        or select(
            "Please select the type of the template",
            valide_template_types,
            default='generic'
        ).ask()
    )

    if type is None:
        logger.warning('Cancelled by user, exiting...')
        exit()

    if type not in valide_template_types:
        logger.error(
            f"Invalid template type '{template_type}'. Valid types are: {', '.join(valide_template_types)}"
        )
        return

    # -----------------------------------------------------------------

    if directory and directory[0] != "/":
        directory = os.path.join(os.getcwd(), directory)
    else:
        directory = path("Where should the exploied be saved to?", default=os.path.join(os.getcwd(), f'{service}-{name}'), only_directories=True).ask()
    
    if directory is None:
        logger.warning('Cancelled by user, exiting...')
        exit()

    if os.path.exists(directory) and len(os.listdir(directory)) > 0:
        logger.warning(f"The Path '{directory}' is not empty!")

        if confirm("Do you want to overwrite the directory?").ask():
            shutil.rmtree(directory)
        else:
            logger.info("Directory won't be overwritten, exiting...")
            exit()
    
    logger.info(f"setting up directory: {directory}")
    logger.info("requesting template")

    r = request.post(
        f"api/services/{service}/template/",
        params={"exploit_name": name, "exploit_author": author, "type":type, "additional_args": ""},
    )
    if not r.ok:
        raise click.ClickException(
            f"Requesting template from remote failed with error: {r.text}. Did you run novara configure?"
        )

    logger.info("extracting template")
    zip_template = BytesIO(r.content)
    os.mkdir(directory)
    
    with ZipFile(zip_template) as zip:
        for zip_info in zip.infolist():
            extracted_path = os.path.join(directory, zip_info.filename)
            if os.path.exists(extracted_path):
                if not confirm(f"'{zip_info.filename}' already exists. Do you want to overwrite it?", default=False).ask():
                    continue
            zip.extract(zip_info, directory)

    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".py"):
                os.chmod(os.path.join(root, file), 0o755)

    logger.info(f"Template extracted sucessfully into directory {directory or f'{service}-{name}'}")
    logger.info("To add a new dependency run 'novara generate'")
    logger.info("To run the current exploit run 'novara run [shell|local|remote]'")
    logger.info("Happy exploiting!")



import requests
import webbrowser
import time

from pydantic import BaseModel
from typing import Optional
from urllib.parse import urljoin

from novara.config import config, ConfigManager, Token

from requests.exceptions import JSONDecodeError

class DeviceCodeResponse(BaseModel):
    device_code: str
    verification_uri: str
    verification_uri_complete: str
    user_code: str
    expires_in: int
    interval: int
    
class UserinfoModel(BaseModel):
    sub:str
    email:str
    email_verified:bool
    name:str
    given_name:str
    preferred_username:str
    nickname:str
    groups:list[str]

class AuthSession(requests.Session):
    userinfo: Optional[UserinfoModel] = None

    def __init__(self, config:ConfigManager):
        self.config = config
        super().__init__()

    def _request_token(self, from_user_info:bool = False) -> Optional[Token]:
        if self.config.token.is_valid() and not from_user_info and self.get_userinfo():
            return 

        if not self.config.idp_url or not self.config.client_id:
            raise Exception("IDP URL or Client ID not set in configuration")

        device_code_resp = requests.post(
            urljoin(self.config.idp_url, '/application/o/device/'),
            data={'client_id': self.config.client_id, 'scope': 'openid profile email'}, 
            allow_redirects=True
        )
        device_code_resp.raise_for_status()

        device_code = DeviceCodeResponse.model_validate(device_code_resp.json())

        print(f'Please visit {device_code.verification_uri} and enter code {device_code.user_code} or {device_code.verification_uri_complete}')
        webbrowser.open(device_code.verification_uri_complete)

        poll_count = device_code.expires_in // device_code.interval

        for _ in range(poll_count):
            token_resp = requests.post(
                urljoin(self.config.idp_url, '/application/o/token/'),
                data={
                    'client_id': self.config.client_id,
                    'device_code': device_code.device_code,
                    'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
                }
            )

            if token_resp.status_code == 400 and token_resp.json().get('error') == 'authorization_pending':
                time.sleep(device_code.interval)
                continue

            token_resp.raise_for_status()
            self.config.token = Token.model_validate(token_resp.json())
            self.config.save()
            break
        else:
            print('Authentication expired please try again')
            exit()

        if not self.get_userinfo():
            raise Exception("failed to retreive valide userinfo")

        return self.config.token
        
    def get_userinfo(self) -> Optional[UserinfoModel]:
        if not self.config.token.is_valid():
            self._request_token(from_user_info=True)
        
        resp = requests.get(
            urljoin(self.config.idp_url, '/application/o/userinfo/'), 
            headers={
                'Authorization': f'{self.config.token.token_type} {self.config.token.access_token}'
            }
        )
        
        if not resp.ok:
            return None
        
        self.userinfo = UserinfoModel.model_validate(resp.json())

        return self.userinfo
    
    def request(self, method, url:str, params = None, data = None, headers:Optional[dict[str, str]] = None, cookies = None, files = None, auth = None, timeout = None, allow_redirects = True, proxies = None, hooks = None, stream = None, verify = None, cert = None, json = None):
        if not self.config.token.is_valid():
            self._request_token()

        headers = (headers or {}) | {'Authorization': f'{self.config.token.token_type} {self.config.token.access_token}'}

        if not url.startswith(('http://', 'https://')):
            url = urljoin(self.config.server_url, url)
        
        return super().request(method, url, params, data, headers, cookies, files, auth, timeout, allow_redirects, proxies, hooks, stream, verify, cert, json)


request = AuthSession(config)


from pathlib import Path
import os
from importlib.metadata import version

__version__ = version('novara')

CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
CONFIG_FILE = CONFIG_HOME / "novara" / "config.yml"
SSHKEY_FILE = CONFIG_HOME / "novara" / "novara.key"
SOCKET_FILE = CONFIG_HOME / "novara" / "novara_docker.sock"


from pydantic import BaseModel, Field
import yaml
from typing import Union, Optional, Literal
import logging
from datetime import datetime, timedelta

from novara.constants import CONFIG_FILE, CONFIG_HOME

logger = logging.getLogger("rich")


class Token(BaseModel):
    access_token: str = ''
    token_type: str = ''
    scope: str = ''
    expires_in: int = 0
    id_token: str = ''
    created_at:datetime = Field(default_factory=datetime.now)

    class Config:
        extra = 'ignore'

    @property
    def valide_until(self):
        return self.created_at + timedelta(seconds=self.expires_in)
    
    def is_valid(self):
        return self.access_token and self.token_type and self.valide_until > datetime.now()


class Config_Model(BaseModel):
    author:str = ''

    ssh_port:int = 22
    ssh_user:str = ''
    ssh_url:str = ''
    ssh_privatekey:str = ''

    server_url: str = ''

    idp_url:str = ''
    client_id:str = ''

    logging_level:Literal['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'] = 'NOTSET'

    token:Token = Token()

    class Config:
        extra = 'ignore'


class ConfigManager(Config_Model):
    is_initialized: bool = False

    def __init__(self, *args, **kwargs):
        object.__setattr__(self, 'is_initialized', False)       # avoid triggering loading logic
        if args or kwargs:
            super().__init__(*args, **kwargs)

    def _load(self) -> dict:
        logger.info('loading new config...')
        try:
            with open(CONFIG_FILE, 'r') as config_file:
                return yaml.safe_load(config_file)
        except (FileNotFoundError, OSError):
            logger.error('config file not found or not accessible')
            logger.debug('did you run novara configure?')
            exit()
            
    def _initialize(self):
        super().__init__(**self._load())
        self.is_initialized = True
    
    def raw_write(self, config: dict):
        try:
            if not CONFIG_HOME.exists():
                logger.info(f"creating directory {CONFIG_HOME}")
                CONFIG_HOME.mkdir()
            config_directory = CONFIG_FILE.parent
            if not config_directory.exists():
                logger.info(f"creating directory {config_directory}")
                config_directory.mkdir()
            with open(CONFIG_FILE, 'w') as config_file:
                yaml.dump(config, config_file)
        except OSError:
            logger.error("Couldn't create the config file it's not writable")
            exit()

    def update(self, **kwargs):
        """
        Update the configuration with new values.
        """

        for name, value in kwargs.items():
            if name in Config_Model.model_fields:
                setattr(self, name, value)
            else:
                logger.warning(f"Unknown configuration key: {name}")

        self.raw_write(self.model_dump())

    def save(self):
        self.raw_write(self.model_dump())

    def __getattr__(self, name: str):
        if name in Config_Model.model_fields:
            self._initialize()

        return super().__getattribute__(name)

    @property
    def raw_config(self):
        """Access the config as a dict"""
        if not self.is_initialized:
            self._initialize()
        return self.model_dump()

    @raw_config.setter
    def raw_config(self, value: Union[dict, BaseModel]):
        if isinstance(value, BaseModel):
            value = value.model_dump()

        if self.is_initialized:
            value = {**self.model_dump(), **value}

        super().__init__(**value)

config = ConfigManager()


import rich_click as click
from novara.constants import __version__
from novara.utils import logger, test_ssh_connection
from novara.config import config
from novara.request import request
from novara.constants import SOCKET_FILE
from novara.commands.docker import forward_docker_socket, cleanup_docker_socket, docker
import requests
import time
import subprocess
import os


from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

def get_latest_version():
    url = f"https://pypi.org/pypi/novara/json"
    response = requests.get(url)
    if response.status_code == 200:
        data = response.json()
        latest_version = data['info']['version']
        return latest_version
    else:
        raise Exception(f"Failed to fetch package information: {response.status_code}")

@click.command()
def info():
    logger.debug('fetching version of the cli from pypi...')
    latest_version = get_latest_version()

    logger.debug("check connectivity to the backend...")
    try:
        r = request.get("api/up")
        is_up = r.status_code == 200
    except Exception:
        is_up = False

    if is_up:
        logger.debug("time response of the backend...")
        start_time = time.time()
        r = request.get("api/up")
        time_elapsed = time.time() - start_time
    else:
        time_elapsed = None

    console = Console()

    # Header
    console.rule("[bold cyan]Novara CLI Info[/bold cyan]", style="cyan")

    # Version Info
    version_table = Table(show_header=False, box=None)
    version_table.add_row("Novara CLI Version:", f"[green]{__version__}[/green]")
    version_table.add_row("Latest Available:", f"[green]{latest_version}[/green]")
    console.print(version_table)

    if __version__ != latest_version:
        console.print(
            Panel.fit(
                Text(
                    "You are using an older version of the CLI.\n"
                    "Consider upgrading:\n"
                    "    pip install --upgrade novara",
                    style="yellow bold"
                ),
                border_style="yellow"
            )
        )

    # Backend Info
    backend_table = Table(show_header=False, box=None)
    backend_table.add_row("Backend Server:", f"[blue]{config.server_url}[/blue]")
    backend_table.add_row("Author:", f"{config.author}")
    if is_up and time_elapsed is not None:
        backend_table.add_row("Backend Status:", f"[green]Reachable ({time_elapsed:.2f}s response)[/green]")
    else:
        backend_table.add_row("Backend Status:", "[red]Unreachable[/red]")
    console.print(backend_table)

    console.rule(style="cyan")
